//
// Copyright © 2017 Gavrilov Daniil
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

import QuartzCore
import Combine

/// Memory usage tuple. Contains used and total memory in bytes.
public typealias MemoryUsage = (used: UInt64, total: UInt64)

/// Sample data
public typealias SampleData = (cpu: Double, memory: MemoryUsage)

// MARK: Class Definition

/// Performance calculator. Uses CADisplayLink to count FPS. Also counts CPU and memory usage.
internal class PerformanceCalculator {
    // MARK: Structs
    
    private struct Constants {
		static let accumulationTimeInSeconds = 0.1
    }
    
    private var startTimestamp: TimeInterval?
    private var accumulatedInformationIsEnough = false

	// MARK: Public Methods
	private var timerCancellable: AnyCancellable?

	var pid: Int32 = 0

	public func run(pid: Int32) {
		self.startTimestamp = Date().timeIntervalSince1970
		self.pid = pid

		timerCancellable = Timer.publish(every: 1, tolerance: nil, on: RunLoop.main, in: .common)
			.autoconnect()
			.sink(receiveValue: { _ in
				let data = self.takePerformanceEvidence()
				if data.memory.total > 0 {
					self.usage = data
				}
			})
	}

	@Published var usage: SampleData = (cpu: 0, memory: (used: 0, total: 0))

	public func stop() {
		timerCancellable?.cancel()
		timerCancellable = nil
	}
}

// MARK: Monitoring

private extension PerformanceCalculator {
    func takePerformanceEvidence() -> SampleData {
        if self.accumulatedInformationIsEnough {
            let cpuUsage = self.cpuUsage()
            let memoryUsage = self.memoryUsage()
			return (cpu: cpuUsage, memory: memoryUsage)
		} else if let start = self.startTimestamp, Date().timeIntervalSince1970 - start >= Constants.accumulationTimeInSeconds {
            self.accumulatedInformationIsEnough = true
        }
		return (cpu: 0.0, memory: (used: 0, total: 0))
    }
    
    func cpuUsage() -> Double {
        var totalUsageOfCPU: Double = 0.0
        var threadsList: thread_act_array_t?
        var threadsCount = mach_msg_type_number_t(0)
		var task: mach_port_t = 0

		let taskResult = task_for_pid(mach_task_self_, pid, &task )
		guard taskResult == KERN_SUCCESS else {
			//print("Error: could not get the task for pid '\(pid)'")
			stop()
			return 0
		}

        let threadsResult = withUnsafeMutablePointer(to: &threadsList) {
            return $0.withMemoryRebound(to: thread_act_array_t?.self, capacity: 1) {
                task_threads(task, $0, &threadsCount)
            }
        }

        if threadsResult == KERN_SUCCESS, let threadsList = threadsList {
            for index in 0..<threadsCount {
                var threadInfo = thread_basic_info()
                var threadInfoCount = mach_msg_type_number_t(THREAD_INFO_MAX)
                let infoResult = withUnsafeMutablePointer(to: &threadInfo) {
                    $0.withMemoryRebound(to: integer_t.self, capacity: 1) {
                        thread_info(threadsList[Int(index)], thread_flavor_t(THREAD_BASIC_INFO), $0, &threadInfoCount)
                    }
                }
                
                guard infoResult == KERN_SUCCESS else {
                    break
                }
                
                let threadBasicInfo = threadInfo as thread_basic_info
                if threadBasicInfo.flags & TH_FLAGS_IDLE == 0 {
                    totalUsageOfCPU = (totalUsageOfCPU + (Double(threadBasicInfo.cpu_usage) / Double(TH_USAGE_SCALE) * 100.0))
                }
            }
        }
        
        vm_deallocate(mach_task_self_, vm_address_t(UInt(bitPattern: threadsList)), vm_size_t(Int(threadsCount) * MemoryLayout<thread_t>.stride))
        return totalUsageOfCPU
    }
    
    func memoryUsage() -> MemoryUsage {
		var task: mach_port_t = 0

		let taskResult = task_for_pid(mach_task_self_, pid, &task )
		guard taskResult == KERN_SUCCESS else {
			//print("Error: could not get the task for pid '\(pid)'")
			return (used: 0, total: 0)
		}

        var taskInfo = task_vm_info_data_t()
        var count = mach_msg_type_number_t(MemoryLayout<task_vm_info>.size) / 4
        let result: kern_return_t = withUnsafeMutablePointer(to: &taskInfo) {
            $0.withMemoryRebound(to: integer_t.self, capacity: 1) {
                task_info(task, task_flavor_t(TASK_VM_INFO), $0, &count)
            }
        }
        
        var used: UInt64 = 0
        if result == KERN_SUCCESS {
            used = UInt64(taskInfo.phys_footprint)
        }
        
        let total = ProcessInfo.processInfo.physicalMemory
        return (used, total)
    }
}
